# package Spawn;

#  This set of subroutines handles forking multiple processes
#
#  spawn	name command-line
#  sp_print	name message
#  sp_wait	
#	    commands source message
#  
#  Each process may be individually written to using "sp_print".
#  The command "sp_wait" reads the stdout of all processes using
#  select; any
#  output is given to the routine "commands", which must be supplied
#  by the user.
# 

use POSIX ":sys_wait_h";
use rt ;
use Datascope ; 
use Carp qw(cluck) ;

sub sp_wait {
    my($timeout) = 0.5 ;
    my($nbytes, $i, $ready, $n, $FILEHANDLE, $input, $name, $line ) ; 
    my ( $death ) ;
    while ( 1 ) {
	$death = 0 ; 
	my ( $nfound, $timeleft ) = 
	    select ( $reads=$Readfds, 
		     $writes=$Writefds, 
		     $exceptions=$Readfds, 
		     $timeout )  ;
	if ( $nfound ) { 
	    $ready = unpack ( "b*", $reads ) ; 
	    $n = length($ready) ; 
	    for ( $i=0 ; $i<$n ; $i++ ) {
		if ( substr($ready, $i, 1) == '1' ) {
		    $name = $Sp_Name{$i} ; 
		    $FILEHANDLE = $Sp_Read{$name} ; 
		    $input = "" ; # $Sp_buf{$name} ; 
		    $nbytes = sysread ($FILEHANDLE, $input, 8192 ) ; 
		    if ( $nbytes < 1 || ! defined $nbytes ) {
			$death = 1 ; 
		    } else {
			foreach $line ( split ( '\n', $input ) ) {
			    &commands ( $name, $line ) ; 
			}
		    }
		}
	    }
	}

	( $nfound, $timeleft ) = 
	    select ( $reads=$Errfds,
		     $writes="",
		     $exceptions="",
		     $timeout )  ;
	$nfound = 0 ; 
	if ( $nfound ) { 
	    $ready = unpack ( "b*", $reads ) ; 
	    $n = length($ready) ; 
	    print STDERR "nfound = $nfound, ready=$ready, n=$n\n" ;
	    for ( $i=0 ; $i<$n ; $i++ ) {
		if ( substr($ready, $i, 1) == '1' ) {
		    $name = $Sp_Name{$i} ; 
		    $FILEHANDLE = $Sp_ErrRead{$name} ; 
		    $input = "" ; # $Sp_buf{$name} ; 
		    $nbytes = sysread ($FILEHANDLE, $input, 8192 ) ; 
		    if ( $nbytes < 1 || ! defined $nbytes ) {
			print STDERR "\nDid '$Sp_Cmd{$name}' die?\n" ; 
			$death = 1 ; 
		    } else {
			print STDERR "\nErrors from '$Sp_Cmd{$name}'\n" ; 
			print STDERR $input ; 
		    }
		    print STDERR "\n" ; 
		}
	    }
	} else { # idle loop
	}
	if ( $death ) { 
	    &death_of_child() ; 
	}
    }
}

sub spawn {  # spawn ( name, cmd ) 
    my ( $name, $cmd ) = @_ ; 
    if ( $cmd =~ /^<\s*(.*)/ ) { 
	my $filename = $1 ;
	my $pid = -1 ;
	$Sp_Name{$pid} = $name ; 
	$Sp_Cmd{$name} = $cmd ;
	$Sp_Pid{$name} = $pid ; 
	if ( -p $filename ) { 
	    system ( "echo '' > $filename &" ) ; 
	}
	my $PREAD = "pr$name" ; 
	sysopen ($PREAD, $filename, 0 ) or  # 0 means O_RDONLY
	    carp ( "Can't open $filename : $!\n" ) ;
	$Sp_Read{$name} = $PREAD ; 
	$fileno = fileno($PREAD) ; 
	$Sp_Name{$fileno} = $name ; 
	vec ( $Readfds, $fileno, 1) = 1; 
	return ( $pid, $PREAD ) ;   
    } else {
	my ( $pid, $fileno ) ;
	my ( $CREAD, $CWRITE, $PREAD, $PWRITE, $PERRREAD, $CERRWRITE ) ;

	$CREAD = "cr$name" ; 
	$CWRITE = "cw$name" ; 
	$PREAD = "pr$name" ; 
	$PWRITE = "pw$name" ; 
	$PERRREAD = "pre$name" ; 
	$CERRWRITE = "cwe$name" ; 

	pipe ( $CREAD, $PWRITE ) ;
	pipe ( $PREAD, $CWRITE ) ; 
	pipe ( $PERRREAD, $CERRWRITE ) ; 
	if ( $pid = fork ) { 
	    # parent
	    close $CWRITE ; 
	    close $CREAD ; 
	    close $CERRWRITE ;
	    $Sp_Name{$pid} = $name ; 
	    $Sp_Cmd{$name} = $cmd ;
	    $Sp_Pid{$name} = $pid ; 
	    $Sp_Read{$name} = $PREAD ; 
	    $Sp_Write{$name} = $PWRITE ; 
	    $Sp_ErrRead{$name} = $PERRREAD ; 
	    $fileno = fileno($PREAD) ; 
	    $Sp_Name{$fileno} = $name ; 
	    vec ( $Readfds, $fileno, 1) = 1; 
	    $fileno = fileno($PERRREAD) ; 
	    $Sp_Name{$fileno} = $name ; 
	    vec ( $Errfds, $fileno, 1) = 1; 
	    select ((select($PWRITE), $|=1) [0] ) ; 
	    return ( $pid, $PREAD, $PWRITE ) ;   
	} elsif ( defined $pid ) {
	    # child 
	    close $PWRITE ; 
	    close $PREAD ; 
	    close $PERRREAD ; 
	    #$SIG{'TERM'} = 'IGNORE' ; 
	    $SIG{'STOP'} = 'IGNORE' ; 
	    $SIG{'TSTP'} = 'IGNORE' ; 
	    # $SIG{'CHLD'} = 'IGNORE' ; 
	    if ( $] > 5 ) { 
		eval "setpgrp 0" ; 
	    } else { 
		setpgrp (0, $$) ; 
	    }
	    open ( STDIN, "<&$CREAD" ) ; 
	    open ( STDOUT, ">&$CWRITE" ) ; 
	    open ( STDERR, ">/dev/tty" ) ; 
	    close $CERRWRITE ;
	    # open ( STDERR, ">&$CERRWRITE" ) ;  # this somehow flummoxes dbloc2

	    exec $cmd ; 
	    die "Cannot exec $cmd: $!" ; 
	} else {
	    die "Can't fork $cmd\n" ; 
	}
    }
}


sub sp_print { 
    my ( $name, $line ) = @_ ; 
    
    my ( $FILEHANDLE ) = $Sp_Write{$name} ; 
    if ( defined $FILEHANDLE ) { 
	if ( defined $line ) { 
	    print $FILEHANDLE "$line\n" ; 
	} else { 
	    print $FILEHANDLE "\n" ;
	    # cluck "line not defined in sp_print!" ; 
	}
	# print STDERR "sending '$line' to $name\n" ; 
    } else { 
	print STDERR "No filehandle for '$name' to send '$line'\n" ; 
    }
}

sub death_of_child { 
    my ( $pid, $status, $name, $sig, $result ) ; 
    my ( $dumped_core, $FILEHANDLE, $input, $nbytes, $msg, $fileno );
    if ( $Norestart ) { return ; } 
    while (1) {
        $pid = waitpid(-1,&WNOHANG);
        last if ($pid < 1);
	$status = $? ; 
	if ( ! defined $Sp_Name{$pid} ) { 
	    printf STDERR ">> Unspawned child $pid died with status %02x.\n", $status ; 
	    return ;
	} 
	$name = $Sp_Name{$pid} ; 

	$FILEHANDLE = $Sp_ErrRead{$name} ; 
	$input = "" ; # $Sp_buf{$name} ; 
	$nbytes = sysread ($FILEHANDLE, $input, 8192 ) ; 
	if ( $nbytes > 0 ) { 
	    print STDERR "\nErrors from '$Sp_Cmd{$name}'\n" ; 
	    print "\nErrors from '$Sp_Cmd{$name}'\n" ; 
	    while ( $nbytes > 0 ) { 
		print STDERR $input ;
		print "$input" ; 
		$input = "" ;
		$nbytes = sysread ($FILEHANDLE, $input, 8192 ) ; 
	    }
	}

	$sig = ( $status & 127 ) ; 
	$dumped_core = $status & 128 ;
	$result = $status >> 8 ;
	print STDERR "death_of_child: sig=$sig dumped_core=$dumped_core result=$result\n" ; 
	if ( $sig != 0 ) {
	    $msg = sprintf(
		    "'%s' (%s) died from signal %s: %s with exit code %d\n",
		    $Sp_Cmd{$name}, $pid, $sig, strsignal($sig), $result) ;
	} else {
	    $msg = sprintf ( "'%s' (%s) died with exit code %d\n",
		    $Sp_Cmd{$name}, $pid, $result ) ;
	}

	print STDERR $msg ;
	print $msg ;

	$restart++ ;
	if ( defined &my_death_of_child ) {
	    &my_death_of_child ( $name, $pid, $sig, $result, $dumped_core  ) ; 
	}
	if ( $restart > 2 ) {
	    print STDERR "Quitting because of repeated failures.\n" ;
	    print "Quitting because of repeated failures.\n" ;
	    &quit ;
	}

	$FILEHANDLE =  $Sp_Read{$name} ;
	$fileno = fileno($FILEHANDLE) ; 
	vec ( $Readfds, $fileno, 1) = 0; 
	close $FILEHANDLE ;

	$FILEHANDLE =  $Sp_Write{$name} ;
	close $FILEHANDLE ;

	$FILEHANDLE = $Sp_ErrRead{$name} ;
	$fileno = fileno($FILEHANDLE) ; 
	vec ( $Errfds, $fileno, 1) = 0; 
	close $FILEHANDLE ;

	print "Attempting to restart $name.\n" ;
	print STDERR "Attempting to restart $name.\n\n" ;
	&spawn ( $name, $Sp_Cmd{$name} ) ;
    }
}


sub sp_die { 
    my ($pid, $file, @pids, $i) ; 
    @pids = values %Sp_Pid ;
    @pids = grep($_ > 0, @pids) ;
    print "Terminating all sub processes:\n\t@pids\n" ; 
    foreach $pid ( @pids ) { 
	printf "    %5d %s\n", $pid, $Sp_Name{$pid} ;
    }
    $Norestart = 1 ;
    my @msgs = &rtkill ( 10, @pids ) ;
    #print "Terminated all sub processes:\n\t@pids\n" ; 
    #print "    ", join("\n    ", @msgs), "\n" ;
}

$Writefds  = '' ;
$Readfds = '' ;
$Errfds = '' ; 
    
1;

# $Id$ 
